from typing import Optional

from pydantic import AnyUrl, BaseSettings, Field


class WebSocketUrl(AnyUrl):
    allowed_schemes = {"wss", "ws"}


class Settings(BaseSettings):
    class Config:
        env_prefix = ""  # defaults to no prefix, i.e. ""
        fields = {
            "redis_dns": {"env": ["redishost", "redis_host", "redis_dns"]},
            "redis_password": {"env": ["redis_password", "redispass"]},
        }

    application_name: str = Field(default="pyetp")
    application_version: str = Field(default="0.0.33")

    # dataspace: str = Field(default='demo/pss-data-gateway')
    etp_url: WebSocketUrl = Field(default="wss://host.com")
    etp_timeout: float = Field(default=15.0, description="Timeout in seconds")
    data_partition: Optional[str] = None
    MaxWebSocketMessagePayloadSize: int = Field(default=500000000)


SETTINGS = Settings()



import asyncio
import datetime
import logging
import sys
import typing as T
import uuid
import warnings
from collections import defaultdict
from types import TracebackType

import numpy as np
import websockets
from etpproto.connection import CommunicationProtocol, ConnectionType, ETPConnection
from etpproto.messages import Message, MessageFlags
from etptypes import ETPModel
from etptypes.energistics.etp.v12.datatypes.any_array_type import AnyArrayType
from etptypes.energistics.etp.v12.datatypes.any_logical_array_type import (
    AnyLogicalArrayType,
)
from etptypes.energistics.etp.v12.datatypes.array_of_string import ArrayOfString
from etptypes.energistics.etp.v12.datatypes.data_array_types.data_array_identifier import (
    DataArrayIdentifier,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.data_array_metadata import (
    DataArrayMetadata,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.get_data_subarrays_type import (
    GetDataSubarraysType,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.put_data_arrays_type import (
    PutDataArraysType,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.put_data_subarrays_type import (
    PutDataSubarraysType,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.put_uninitialized_data_array_type import (
    PutUninitializedDataArrayType,
)
from etptypes.energistics.etp.v12.datatypes.data_value import DataValue
from etptypes.energistics.etp.v12.datatypes.error_info import ErrorInfo
from etptypes.energistics.etp.v12.datatypes.object.context_info import ContextInfo
from etptypes.energistics.etp.v12.datatypes.object.context_scope_kind import (
    ContextScopeKind,
)
from etptypes.energistics.etp.v12.datatypes.object.data_object import DataObject
from etptypes.energistics.etp.v12.datatypes.object.dataspace import Dataspace
from etptypes.energistics.etp.v12.datatypes.object.relationship_kind import (
    RelationshipKind,
)
from etptypes.energistics.etp.v12.datatypes.object.resource import Resource
from etptypes.energistics.etp.v12.datatypes.supported_data_object import (
    SupportedDataObject,
)
from etptypes.energistics.etp.v12.datatypes.supported_protocol import SupportedProtocol
from etptypes.energistics.etp.v12.datatypes.uuid import Uuid
from etptypes.energistics.etp.v12.datatypes.version import Version
from etptypes.energistics.etp.v12.protocol.core.authorize import Authorize
from etptypes.energistics.etp.v12.protocol.core.authorize_response import (
    AuthorizeResponse,
)
from etptypes.energistics.etp.v12.protocol.core.close_session import CloseSession
from etptypes.energistics.etp.v12.protocol.core.open_session import OpenSession
from etptypes.energistics.etp.v12.protocol.core.protocol_exception import (
    ProtocolException,
)
from etptypes.energistics.etp.v12.protocol.core.request_session import RequestSession
from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata import (
    GetDataArrayMetadata,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata_response import (
    GetDataArrayMetadataResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays import (
    GetDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays_response import (
    GetDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays import (
    GetDataSubarrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays_response import (
    GetDataSubarraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays import (
    PutDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays_response import (
    PutDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays import (
    PutDataSubarrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays_response import (
    PutDataSubarraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays import (
    PutUninitializedDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays_response import (
    PutUninitializedDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.dataspace.delete_dataspaces import (
    DeleteDataspaces,
)
from etptypes.energistics.etp.v12.protocol.dataspace.delete_dataspaces_response import (
    DeleteDataspacesResponse,
)
from etptypes.energistics.etp.v12.protocol.dataspace.get_dataspaces import GetDataspaces
from etptypes.energistics.etp.v12.protocol.dataspace.get_dataspaces_response import (
    GetDataspacesResponse,
)
from etptypes.energistics.etp.v12.protocol.dataspace.put_dataspaces import PutDataspaces
from etptypes.energistics.etp.v12.protocol.dataspace.put_dataspaces_response import (
    PutDataspacesResponse,
)
from etptypes.energistics.etp.v12.protocol.discovery.get_resources import GetResources
from etptypes.energistics.etp.v12.protocol.store.delete_data_objects import (
    DeleteDataObjects,
)
from etptypes.energistics.etp.v12.protocol.store.delete_data_objects_response import (
    DeleteDataObjectsResponse,
)
from etptypes.energistics.etp.v12.protocol.store.get_data_objects import GetDataObjects
from etptypes.energistics.etp.v12.protocol.store.get_data_objects_response import (
    GetDataObjectsResponse,
)
from etptypes.energistics.etp.v12.protocol.store.put_data_objects import PutDataObjects
from etptypes.energistics.etp.v12.protocol.store.put_data_objects_response import (
    PutDataObjectsResponse,
)
from etptypes.energistics.etp.v12.protocol.transaction.commit_transaction import (
    CommitTransaction,
)
from etptypes.energistics.etp.v12.protocol.transaction.rollback_transaction import (
    RollbackTransaction,
)
from etptypes.energistics.etp.v12.protocol.transaction.start_transaction import (
    StartTransaction,
)
from pydantic import SecretStr
from xtgeo import RegularSurface

import resqml_objects.v201 as ro
from pyetp import utils_arrays, utils_xml
from pyetp.config import SETTINGS
from pyetp.uri import DataObjectURI, DataspaceURI
from resqml_objects import parse_resqml_v201_object, serialize_resqml_v201_object

logger = logging.getLogger(__name__)

try:
    # for py >3.11, we can raise grouped exceptions
    from builtins import ExceptionGroup  # type: ignore
except ImportError:
    # Python 3.10
    def ExceptionGroup(msg, errors):
        return errors[0]


try:
    # Python >= 3.11
    from asyncio import timeout
except ImportError:
    # Python 3.10
    from contextlib import asynccontextmanager

    import async_timeout

    @asynccontextmanager
    async def timeout(delay: T.Optional[float]) -> T.Any:
        try:
            async with async_timeout.timeout(delay):
                yield None
        except asyncio.CancelledError as e:
            raise asyncio.TimeoutError(f"Timeout ({delay}s)") from e


class ETPError(Exception):
    def __init__(self, message: str, code: int):
        self.message = message
        self.code = code
        super().__init__(f"{message} ({code=:})")

    @classmethod
    def from_proto(cls, error: ErrorInfo):
        assert error is not None, "passed no error info"
        return cls(error.message, error.code)

    @classmethod
    def from_protos(cls, errors: T.Iterable[ErrorInfo]):
        return list(map(cls.from_proto, errors))


def get_all_etp_protocol_classes():
    """Update protocol - all exception protocols are now per message"""

    pddict = ETPConnection.generic_transition_table
    # protocol exception
    pexec = ETPConnection.generic_transition_table["0"]["1000"]

    for v in pddict.values():
        v["1000"] = pexec

    return pddict


class ETPClient(ETPConnection):
    generic_transition_table = get_all_etp_protocol_classes()

    _recv_events: T.Dict[int, asyncio.Event]
    _recv_buffer: T.Dict[int, T.List[ETPModel]]

    def __init__(self, ws: websockets.ClientConnection, timeout=10.0):
        super().__init__(connection_type=ConnectionType.CLIENT)
        self._recv_events = {}
        self._recv_buffer = defaultdict(lambda: list())  # type: ignore
        self.ws = ws

        self.timeout = timeout
        self.client_info.endpoint_capabilities["MaxWebSocketMessagePayloadSize"] = (
            SETTINGS.MaxWebSocketMessagePayloadSize
        )
        self.__recvtask = asyncio.create_task(self.__recv__())

    #
    # client
    #

    async def send(self, body: ETPModel):
        correlation_id = await self._send(body)
        return await self._recv(correlation_id)

    async def _send(self, body: ETPModel):
        msg = Message.get_object_message(body, message_flags=MessageFlags.FINALPART)
        if msg is None:
            raise TypeError(f"{type(body)} not valid etp protocol")

        msg.header.message_id = self.consume_msg_id()
        logger.debug(f"sending {msg.body.__class__.__name__} {repr(msg.header)}")

        # create future recv event
        self._recv_events[msg.header.message_id] = asyncio.Event()

        for msg_part in msg.encode_message_generator(self.max_size, self):
            await self.ws.send(msg_part)

        return msg.header.message_id

    async def _recv(self, correlation_id: int) -> ETPModel:
        assert correlation_id in self._recv_events, (
            "trying to recv response on non-existing message"
        )

        async with timeout(self.timeout):
            await self._recv_events[correlation_id].wait()

        # cleanup
        bodies = self._clear_msg_on_buffer(correlation_id)

        # error handling
        errors = self._parse_error_info(bodies)

        if len(errors) == 1:
            raise ETPError.from_proto(errors.pop())
        elif len(errors) > 1:
            raise ExceptionGroup(
                "Server responded with ETPErrors:", ETPError.from_protos(errors)
            )

        if len(bodies) > 1:
            logger.warning(f"Recived {len(bodies)} messages, but only expected one")

        # ok
        return bodies[0]

    @staticmethod
    def _parse_error_info(bodies: list[ETPModel]) -> list[ErrorInfo]:
        # returns all error infos from bodies
        errors = []
        for body in bodies:
            if isinstance(body, ProtocolException):
                if body.error is not None:
                    errors.append(body.error)
                errors.extend(body.errors.values())
        return errors

    async def close(self, reason=""):
        try:
            await self._send(CloseSession(reason=reason))
        except websockets.ConnectionClosed:
            logger.error(
                "Websockets connection is closed, unable to send a CloseSession-message"
                " to the server"
            )
        finally:
            # Check if the receive task is done, and if not, stop it.
            if not self.__recvtask.done():
                self.__recvtask.cancel("stopped")

            self.is_connected = False

        try:
            # Raise any potential exceptions that might have occured in the
            # receive task
            await self.__recvtask
        except asyncio.CancelledError:
            # No errors except for a cancellation, which is to be expected.
            pass
        except websockets.ConnectionClosed as e:
            # The receive task errored on a closed websockets connection.
            logger.error(
                "The receiver task errored on a closed websockets connection. The "
                f"message was: {e.__class__.__name__}: {e}"
            )

        if len(self._recv_buffer) > 0:
            logger.error(
                f"Connection is closed, but there are {len(self._recv_buffer)} "
                "messages left in the buffer"
            )

        # Check if there were any messages left in the websockets connection.
        # Reading them will speed up the closing of the connection.
        counter = 0
        try:
            async for msg in self.ws:
                counter += 1
        except websockets.ConnectionClosed:
            # The websockets connection had already closed. Either successfully
            # or with an error, but we ignore both cases.
            pass

        if counter > 0:
            logger.error(
                f"There were {counter} unread messages in the websockets connection "
                "after the session was closed"
            )

        logger.debug("Client closed")

    #
    #
    #

    def _clear_msg_on_buffer(self, correlation_id: int):
        del self._recv_events[correlation_id]
        return self._recv_buffer.pop(correlation_id)

    def _add_msg_to_buffer(self, msg: Message):
        self._recv_buffer[msg.header.correlation_id].append(msg.body)

        # NOTE: should we add task to autoclear buffer message if never waited on ?
        if msg.is_final_msg():
            # set response on send event
            self._recv_events[msg.header.correlation_id].set()

    async def __recv__(self):
        logger.debug("starting recv loop")

        while True:
            msg_data = await self.ws.recv()
            msg = Message.decode_binary_message(
                T.cast(bytes, msg_data), ETPClient.generic_transition_table
            )

            if msg is None:
                logger.error(f"Could not parse {msg_data}")
                continue

            logger.debug(f"recv {msg.body.__class__.__name__} {repr(msg.header)}")
            self._add_msg_to_buffer(msg)

    #
    # session related
    #

    async def request_session(self):
        # Handshake protocol
        etp_version = Version(major=1, minor=2, revision=0, patch=0)

        def get_protocol_server_role(protocol: CommunicationProtocol) -> str:
            match protocol:
                case CommunicationProtocol.CORE:
                    return "server"
                case CommunicationProtocol.CHANNEL_STREAMING:
                    return "producer"

            return "store"

        msg = await self.send(
            RequestSession(
                applicationName=SETTINGS.application_name,
                applicationVersion=SETTINGS.application_version,
                clientInstanceId=uuid.uuid4(),  # type: ignore
                requestedProtocols=[
                    SupportedProtocol(
                        protocol=p.value,
                        protocolVersion=etp_version,
                        role=get_protocol_server_role(p),
                    )
                    for p in CommunicationProtocol
                ],
                supportedDataObjects=[
                    SupportedDataObject(qualifiedType="resqml20.*"),
                    SupportedDataObject(qualifiedType="eml20.*"),
                ],
                currentDateTime=self.timestamp,
                earliestRetainedChangeTime=0,
                endpointCapabilities=dict(
                    MaxWebSocketMessagePayloadSize=DataValue(item=self.max_size)
                ),
            )
        )
        assert msg and isinstance(msg, OpenSession)

        self.is_connected = True

        # ignore this endpoint
        _ = msg.endpoint_capabilities.pop("MessageQueueDepth", None)
        self.client_info.negotiate(msg)

        return self

    async def authorize(
        self, authorization: str, supplemental_authorization: T.Mapping[str, str] = {}
    ):
        msg = await self.send(
            Authorize(
                authorization=authorization,
                supplementalAuthorization=supplemental_authorization,
            )
        )
        assert msg and isinstance(msg, AuthorizeResponse)

        return msg

    #

    @property
    def max_size(self):
        return SETTINGS.MaxWebSocketMessagePayloadSize
        # return self.client_info.getCapability("MaxWebSocketMessagePayloadSize")

    @property
    def max_array_size(self):
        return self.max_size - 512  # maxsize - 512 bytes for header and body

    @property
    def timestamp(self):
        return int(datetime.datetime.now(datetime.timezone.utc).timestamp())

    def dataspace_uri(self, ds: str) -> DataspaceURI:
        if ds.count("/") > 1:
            raise Exception("Max one / in dataspace name")
        return DataspaceURI.from_name(ds)

    def list_objects(self, dataspace_uri: DataspaceURI, depth: int = 1) -> list:
        return self.send(
            GetResources(
                scope=ContextScopeKind.TARGETS_OR_SELF,
                context=ContextInfo(
                    uri=dataspace_uri.raw_uri,
                    depth=depth,
                    dataObjectTypes=[],
                    navigableEdges=RelationshipKind.PRIMARY,
                ),
            )
        )

    #
    # dataspace
    #

    async def get_dataspaces(
        self, store_last_write_filter: int = None
    ) -> GetDataspacesResponse:
        return await self.send(
            GetDataspaces(store_last_write_filter=store_last_write_filter)
        )

    async def put_dataspaces(
        self,
        legaltags: list[str],
        otherRelevantDataCountries: list[str],
        owners: list[str],
        viewers: list[str],
        *dataspace_uris: DataspaceURI,
    ):
        _uris = list(map(DataspaceURI.from_any, dataspace_uris))
        for i in _uris:
            if i.raw_uri.count("/") > 4:  # includes the 3 eml
                raise Exception("Max one / in dataspace name")
        time = self.timestamp
        response = await self.send(
            PutDataspaces(
                dataspaces={
                    d.raw_uri: Dataspace(
                        uri=d.raw_uri,
                        storeCreated=time,
                        storeLastWrite=time,
                        path=d.dataspace,
                        custom_data={
                            "legaltags": DataValue(
                                item=ArrayOfString(values=legaltags)
                            ),
                            "otherRelevantDataCountries": DataValue(
                                item=ArrayOfString(values=otherRelevantDataCountries)
                            ),
                            "owners": DataValue(item=ArrayOfString(values=owners)),
                            "viewers": DataValue(item=ArrayOfString(values=viewers)),
                        },
                    )
                    for d in _uris
                }
            )
        )
        assert isinstance(response, PutDataspacesResponse), (
            "Expected PutDataspacesResponse"
        )

        assert len(response.success) == len(dataspace_uris), (
            f"expected {len(dataspace_uris)} success's"
        )

        return response.success

    async def put_dataspaces_no_raise(
        self,
        legaltags: list[str],
        otherRelevantDataCountries: list[str],
        owners: list[str],
        viewers: list[str],
        *dataspace_uris: DataspaceURI,
    ):
        try:
            return await self.put_dataspaces(
                legaltags, otherRelevantDataCountries, owners, viewers, *dataspace_uris
            )
        except ETPError:
            pass

    async def delete_dataspaces(self, *dataspace_uris: DataspaceURI):
        _uris = list(map(str, dataspace_uris))

        response = await self.send(DeleteDataspaces(uris=dict(zip(_uris, _uris))))
        assert isinstance(response, DeleteDataspacesResponse), (
            "Expected DeleteDataspacesResponse"
        )
        return response.success

    #
    # data objects
    #

    async def get_data_objects(self, *uris: T.Union[DataObjectURI, str]):
        _uris = list(map(str, uris))

        msg = await self.send(GetDataObjects(uris=dict(zip(_uris, _uris))))
        assert isinstance(msg, GetDataObjectsResponse), "Expected dataobjectsresponse"
        assert len(msg.data_objects) == len(_uris), (
            "Here we assume that all three objects fit in a single record"
        )

        return [msg.data_objects[u] for u in _uris]

    async def put_data_objects(self, *objs: DataObject):
        response = await self.send(
            PutDataObjects(
                data_objects={f"{p.resource.name} - {p.resource.uri}": p for p in objs},
            )
        )

        assert isinstance(response, PutDataObjectsResponse), (
            "Expected PutDataObjectsResponse"
        )

        return response.success

    async def get_resqml_objects(
        self, *uris: T.Union[DataObjectURI, str]
    ) -> T.List[ro.AbstractObject]:
        data_objects = await self.get_data_objects(*uris)
        return [
            parse_resqml_v201_object(data_object.data) for data_object in data_objects
        ]

    async def put_resqml_objects(
        self, *objs: ro.AbstractObject, dataspace_uri: DataspaceURI
    ):
        time = self.timestamp
        uris = [DataObjectURI.from_obj(dataspace_uri, obj) for obj in objs]
        dobjs = [
            DataObject(
                format="xml",
                data=serialize_resqml_v201_object(obj),
                resource=Resource(
                    uri=uri.raw_uri,
                    name=obj.citation.title if obj.citation else obj.__class__.__name__,
                    lastChanged=time,
                    storeCreated=time,
                    storeLastWrite=time,
                    activeStatus="Inactive",  # type: ignore
                    sourceCount=None,
                    targetCount=None,
                ),
            )
            for uri, obj in zip(uris, objs)
        ]

        _ = await self.put_data_objects(*dobjs)
        return uris

    async def delete_data_objects(
        self, *uris: T.Union[DataObjectURI, str], prune_contained_objects=False
    ):
        _uris = list(map(str, uris))

        response = await self.send(
            DeleteDataObjects(
                uris=dict(zip(_uris, _uris)),
                prune_contained_objects=prune_contained_objects,
            )
        )
        assert isinstance(response, DeleteDataObjectsResponse), (
            "Expected DeleteDataObjectsResponse"
        )

        return response.deleted_uris

    async def start_transaction(
        self, dataspace_uri: DataspaceURI, read_only: bool = True
    ) -> Uuid:
        trans_id = await self.send(
            StartTransaction(
                read_only=read_only, dataspace_uris=[dataspace_uri.raw_uri]
            )
        )
        if trans_id.successful is False:
            raise Exception(f"Failed starting transaction {dataspace_uri.raw_uri}")
        # uuid.UUID(bytes=trans_id.transaction_uuid)
        return Uuid(trans_id.transaction_uuid)

    async def commit_transaction(self, transaction_uuid: Uuid):
        r = await self.send(CommitTransaction(transaction_uuid=transaction_uuid))
        if r.successful is False:
            raise Exception(r.failure_reason)
        return r

    async def rollback_transaction(self, transaction_id: Uuid):
        return await self.send(RollbackTransaction(transactionUuid=transaction_id))

    #
    # xtgeo
    #
    async def get_xtgeo_surface(
        self,
        epc_uri: T.Union[DataObjectURI, str],
        gri_uri: T.Union[DataObjectURI, str],
        crs_uri: T.Union[DataObjectURI, str],
    ):
        (
            gri,
            crs,
        ) = await self.get_resqml_objects(gri_uri, crs_uri)
        rotation = crs.areal_rotation.value
        # some checks

        assert isinstance(gri, ro.Grid2dRepresentation), (
            "obj must be Grid2DRepresentation"
        )
        sgeo = gri.grid2d_patch.geometry.points.supporting_geometry  # type: ignore
        if sys.version_info[1] != 10:
            assert isinstance(
                gri.grid2d_patch.geometry.points, ro.Point3dZValueArray
            ), "Points must be Point3dZValueArray"
            assert isinstance(
                gri.grid2d_patch.geometry.points.zvalues, ro.DoubleHdf5Array
            ), "Values must be DoubleHdf5Array"
            assert isinstance(
                gri.grid2d_patch.geometry.points.supporting_geometry,
                ro.Point3dLatticeArray,
            ), "Points support_geo must be Point3dLatticeArray"
            assert isinstance(sgeo, ro.Point3dLatticeArray), (
                "supporting_geometry must be Point3dLatticeArray"
            )
        assert isinstance(
            gri.grid2d_patch.geometry.points.zvalues.values, ro.Hdf5Dataset
        ), "Values must be Hdf5Dataset"

        # get array
        array = await self.get_array(
            DataArrayIdentifier(
                uri=str(epc_uri),
                pathInResource=gri.grid2d_patch.geometry.points.zvalues.values.path_in_hdf_file,
            )
        )

        return RegularSurface(
            ncol=array.shape[0],
            nrow=array.shape[1],
            xinc=sgeo.offset[0].spacing.value,
            yinc=sgeo.offset[1].spacing.value,
            xori=sgeo.origin.coordinate1,
            yori=sgeo.origin.coordinate2,
            values=array,
            rotation=rotation,
            masked=True,
        )

    async def put_xtgeo_surface(
        self,
        surface: RegularSurface,
        epsg_code: int,
        dataspace_uri: DataspaceURI,
        handle_transaction: bool = True,
    ):
        """Returns (epc_uri, crs_uri, gri_uri).

        If `handle_transaction == True` we start a transaction, and commit it
        after the data has been uploaded. Otherwise, we do not handle the
        transactions at all and assume that the user will start and commit the
        transaction themselves.
        """
        assert surface.values is not None, "cannot upload empty surface"

        if handle_transaction:
            transaction_uuid = await self.start_transaction(
                dataspace_uri, read_only=False
            )

        epc, crs, gri = utils_xml.parse_xtgeo_surface_to_resqml_grid(surface, epsg_code)
        epc_uri, crs_uri, gri_uri = await self.put_resqml_objects(
            epc, crs, gri, dataspace_uri=dataspace_uri
        )

        await self.put_array(
            DataArrayIdentifier(
                uri=epc_uri.raw_uri if isinstance(epc_uri, DataObjectURI) else epc_uri,
                pathInResource=gri.grid2d_patch.geometry.points.zvalues.values.path_in_hdf_file,  # type: ignore
            ),
            surface.values.filled(np.nan).astype(np.float32),
        )

        if handle_transaction:
            await self.commit_transaction(transaction_uuid=transaction_uuid)

        return epc_uri, gri_uri, crs_uri

    #
    # array
    #

    async def get_array_metadata(self, *uids: DataArrayIdentifier):
        response = await self.send(
            GetDataArrayMetadata(dataArrays={i.path_in_resource: i for i in uids})
        )
        assert isinstance(response, GetDataArrayMetadataResponse)

        if len(response.array_metadata) != len(uids):
            raise ETPError(f"Not all uids found ({uids})", 11)

        # return in same order as arguments
        return [response.array_metadata[i.path_in_resource] for i in uids]

    async def get_array(self, uid: DataArrayIdentifier):
        # Check if we can download the full array in one go.
        (meta,) = await self.get_array_metadata(uid)
        if utils_arrays.get_transport_array_size(meta) > self.max_array_size:
            return await self._get_array_chunked(uid)

        response = await self.send(
            GetDataArrays(dataArrays={uid.path_in_resource: uid})
        )
        assert isinstance(response, GetDataArraysResponse), (
            "Expected GetDataArraysResponse"
        )

        arrays = list(response.data_arrays.values())
        return utils_arrays.get_numpy_array_from_etp_data_array(arrays[0])

    async def put_array(
        self,
        uid: DataArrayIdentifier,
        data: np.ndarray,
    ):
        logical_array_type, transport_array_type = (
            utils_arrays.get_logical_and_transport_array_types(data.dtype)
        )
        await self._put_uninitialized_data_array(
            uid,
            data.shape,
            transport_array_type=transport_array_type,
            logical_array_type=logical_array_type,
        )
        # Check if we can upload the full array in one go.
        if data.nbytes > self.max_array_size:
            return await self._put_array_chunked(uid, data)

        response = await self.send(
            PutDataArrays(
                data_arrays={
                    uid.path_in_resource: PutDataArraysType(
                        uid=uid,
                        array=utils_arrays.get_etp_data_array_from_numpy(data),
                    )
                }
            )
        )

        assert isinstance(response, PutDataArraysResponse), (
            "Expected PutDataArraysResponse"
        )
        assert len(response.success) == 1, "expected one success from put_array"

        return response.success

    async def get_subarray(
        self,
        uid: DataArrayIdentifier,
        starts: T.Union[np.ndarray, T.List[int]],
        counts: T.Union[np.ndarray, T.List[int]],
    ):
        starts = np.array(starts).astype(np.int64)
        counts = np.array(counts).astype(np.int64)

        logger.debug(f"get_subarray {starts=:} {counts=:}")

        payload = GetDataSubarraysType(
            uid=uid,
            starts=starts.tolist(),
            counts=counts.tolist(),
        )
        response = await self.send(
            GetDataSubarrays(dataSubarrays={uid.path_in_resource: payload})
        )
        assert isinstance(response, GetDataSubarraysResponse), (
            "Expected GetDataSubarraysResponse"
        )

        arrays = list(response.data_subarrays.values())
        return utils_arrays.get_numpy_array_from_etp_data_array(arrays[0])

    async def put_subarray(
        self,
        uid: DataArrayIdentifier,
        data: np.ndarray,
        starts: T.Union[np.ndarray, T.List[int]],
        counts: T.Union[np.ndarray, T.List[int]],
    ):
        # NOTE: This function assumes that the user (or previous methods) have
        # called _put_uninitialized_data_array.

        # starts [start_X, starts_Y]
        # counts [count_X, count_Y]
        # len = 2 [x_start_index, y_start_index]
        starts = np.array(starts).astype(np.int64)
        counts = np.array(counts).astype(np.int64)  # len = 2
        ends = starts + counts  # len = 2

        slices = tuple(map(lambda se: slice(se[0], se[1]), zip(starts, ends)))
        dataarray = utils_arrays.get_etp_data_array_from_numpy(data[slices])
        payload = PutDataSubarraysType(
            uid=uid,
            data=dataarray.data,
            starts=starts.tolist(),
            counts=counts.tolist(),
        )

        logger.debug(
            f"put_subarray {data.shape=:} {starts=:} {counts=:} "
            f"{dataarray.data.item.__class__.__name__}"
        )

        response = await self.send(
            PutDataSubarrays(dataSubarrays={uid.path_in_resource: payload})
        )
        assert isinstance(response, PutDataSubarraysResponse), (
            "Expected PutDataSubarraysResponse"
        )
        assert len(response.success) == 1, "expected one success"
        return response.success

    #
    # chunked get array - ETP will not chunk response - so we need to do it manually
    #

    def _get_chunk_sizes(
        self, shape, dtype: np.dtype[T.Any] = np.dtype(np.float32), offset=0
    ):
        shape = np.array(shape)

        # capsize blocksize
        # remove 512 bytes for headers and body
        max_items = self.max_array_size / dtype.itemsize
        block_size = np.power(max_items, 1.0 / len(shape))
        block_size = min(2048, int(block_size // 2) * 2)

        assert block_size > 8, "computed blocksize unreasonable small"

        all_ranges = [range(s // block_size + 1) for s in shape]
        indexes = np.array(np.meshgrid(*all_ranges)).T.reshape(-1, len(shape))
        for ijk in indexes:
            starts = ijk * block_size
            if offset != 0:
                starts = starts + offset
            ends = np.fmin(shape, starts + block_size)
            if offset != 0:
                ends = ends + offset
            counts = ends - starts
            if any(counts == 0):
                continue
            yield starts, counts

    async def _get_array_chuncked(self, *args, **kwargs):
        warnings.warn(
            "This function is deprecated and will be removed in a later version of "
            "pyetp. Please use the updated function 'pyetp._get_array_chunked'.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._get_array_chunked(*args, **kwargs)

    async def _get_array_chunked(
        self,
        uid: DataArrayIdentifier,
        offset: int = 0,
        total_count: T.Union[int, None] = None,
    ):
        metadata = (await self.get_array_metadata(uid))[0]
        if len(metadata.dimensions) != 1 and offset != 0:
            raise Exception("Offset is only implemented for 1D array")

        if isinstance(total_count, (int, float)):
            buffer_shape = np.array([total_count], dtype=np.int64)
        else:
            buffer_shape = np.array(metadata.dimensions, dtype=np.int64)
        dtype = utils_arrays.get_dtype_from_any_array_type(
            metadata.transport_array_type
        )
        buffer = np.zeros(buffer_shape, dtype=dtype)
        params = []

        async def populate(starts, counts):
            params.append([starts, counts])
            array = await self.get_subarray(uid, starts, counts)
            ends = starts + counts
            slices = tuple(
                map(lambda se: slice(se[0], se[1]), zip(starts - offset, ends - offset))
            )
            buffer[slices] = array
            return

        _ = await asyncio.gather(
            *[
                populate(starts, counts)
                for starts, counts in self._get_chunk_sizes(buffer_shape, dtype, offset)
            ]
        )

        return buffer

    async def _put_array_chuncked(self, *args, **kwargs):
        warnings.warn(
            "This function is deprecated and will be removed in a later version of "
            "pyetp. Please use the updated function 'pyetp._put_array_chunked'.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self._put_array_chunked(*args, **kwargs)

    async def _put_array_chunked(self, uid: DataArrayIdentifier, data: np.ndarray):
        for starts, counts in self._get_chunk_sizes(data.shape, data.dtype):
            await self.put_subarray(uid, data, starts, counts)

        return {uid.uri: ""}

    async def _put_uninitialized_data_array(
        self,
        uid: DataArrayIdentifier,
        shape: T.Tuple[int, ...],
        transport_array_type: AnyArrayType,
        logical_array_type: AnyLogicalArrayType,
    ):
        payload = PutUninitializedDataArrayType(
            uid=uid,
            metadata=(
                DataArrayMetadata(
                    dimensions=list(shape),  # type: ignore
                    transportArrayType=transport_array_type,
                    logicalArrayType=logical_array_type,
                    storeLastWrite=self.timestamp,
                    storeCreated=self.timestamp,
                )
            ),
        )
        response = await self.send(
            PutUninitializedDataArrays(dataArrays={uid.path_in_resource: payload})
        )
        assert isinstance(response, PutUninitializedDataArraysResponse), (
            "Expected PutUninitializedDataArraysResponse"
        )
        assert len(response.success) == 1, "expected one success"
        return response.success


# define an asynchronous context manager
class connect:
    def __init__(self, authorization: T.Optional[SecretStr] = None):
        self.server_url = SETTINGS.etp_url
        self.authorization = authorization
        self.data_partition = SETTINGS.data_partition
        self.timeout = SETTINGS.etp_timeout

    # ... = await connect(...)

    def __await__(self):
        return self.__aenter__().__await__()

    # async with connect(...) as ...:

    async def __aenter__(self):
        headers = {}
        if isinstance(self.authorization, str):
            headers["Authorization"] = self.authorization
        elif isinstance(self.authorization, SecretStr):
            headers["Authorization"] = self.authorization.get_secret_value()
        if self.data_partition is not None:
            headers["data-partition-id"] = self.data_partition

        self.ws = await websockets.connect(
            self.server_url,
            subprotocols=[ETPClient.SUB_PROTOCOL],  # type: ignore
            additional_headers=headers,
            max_size=SETTINGS.MaxWebSocketMessagePayloadSize,
            ping_timeout=self.timeout,
            open_timeout=None,
        )

        self.client = ETPClient(self.ws, timeout=self.timeout)

        try:
            await self.client.request_session()
        except Exception as e:
            # aexit not called if raised in aenter - so manual cleanup here needed
            await self.client.close("Failed to request session")
            raise e

        return self.client

    # exit the async context manager
    async def __aexit__(self, exc_type, exc: Exception, tb: TracebackType):
        await self.client.close()
        await self.ws.close()



import resqml_objects.v201 as ro

resqml_schema_version = "2.0.1"
common_schema_version = "2.0"


def get_content_type_string(
    obj: ro.AbstractObject,
    resqml_schema_version: str = resqml_schema_version,
    common_schema_version: str = common_schema_version,
) -> str:
    # See Energistics Identifier Specification 4.0 (it is downloaded alongside
    # the RESQML v2.0.1 standard) section 4.1 for an explanation on the format
    # of content_type.

    namespace = getattr(obj.Meta, "namespace", None) or getattr(
        obj.Meta, "target_namespace"
    )

    if namespace == "http://www.energistics.org/energyml/data/resqmlv2":
        return (
            f"application/x-resqml+xml;version={resqml_schema_version};"
            f"type={obj.__class__.__name__}"
        )
    elif namespace == "http://www.energistics.org/energyml/data/commonv2":
        return (
            f"application/x-eml+xml;version={common_schema_version};"
            f"type={obj.__class__.__name__}"
        )

    raise NotImplementedError(
        f"Namespace {namespace} from object {obj} is not supported"
    )


def get_data_object_reference(
    obj: ro.AbstractCitedDataObject,
) -> ro.DataObjectReference:
    content_type = get_content_type_string(obj)

    return ro.DataObjectReference(
        content_type=content_type,
        title=obj.citation.title,
        uuid=obj.uuid,
        version_string=obj.citation.version_string,
    )



import logging
import pathlib
import sys
import zipfile

import h5py
import numpy as np
import numpy.typing as npt

import resqml_objects.v201 as ro_201

from .parsers import parse_resqml_v201_object

logger = logging.getLogger(__name__)


def get_resqml_v201_objects(
    epc_filename: str | pathlib.Path, log_failed_objects: bool = False
) -> list[ro_201.AbstractObject]:
    """This function is fairly brute-force; the function reads all files in the
    provided EPC-file, and tries to parse it as one of the auto-generated
    RESQML v2.0.1 objects. If you are unsure that all objects were included you
    should set `log_failed_objects=True` and see if there are any missed
    RESQML-objects. Objects such as `[Content_Types].xml`, `docProps/core.xml`,
    and anything under the `_rels/`-directory are not RESQML-objects, and will
    be logged as failed.


    Parameters
    ----------
    epc_filename: str | pathlib.Path
        The name of the EPC-file. Typically with an `.epc`-extension.
    log_failed_objects: bool
        Set to `True` in order to log the objects that were not parsed as
        RESQML v2.0.1 objects along with their errors. Default is `False`.


    Returns
    -------
    list[ro_201.AbstractObject]
        A list of the parsed RESQML v2.0.1 objects.


    See Also
    --------
    get_arrays_and_paths_in_hdf_file
    """

    if sys.version_info.major == 3 and sys.version_info.minor == 10:
        raise NotImplementedError(
            "The epc-reader 'get_resqml_v201_objects'-function does not work for "
            "Python 3.10. Consider switching to Python 3.11 or up."
        )

    robjs = []
    fail = {}

    with zipfile.ZipFile(epc_filename, "r") as zf:
        for zi in zf.infolist():
            with zf.open(zi.filename) as f:
                c = f.read()

                try:
                    robjs.append(parse_resqml_v201_object(c))
                except AttributeError as e:
                    fail[zi.filename] = e

    if log_failed_objects:
        logger.info(f"Failed to parse: {fail}")

    return robjs


def get_arrays_and_paths_in_hdf_file(
    epc_hdf_filename: str | pathlib.Path,
) -> dict[str, npt.NDArray[np.bool_ | np.number | np.character]]:
    """In this function we read all arrays and their full paths in an
    HDF5-file. In the context of an EPC-file this corresponds to the
    `pathInHdfFile`-attribute in a RESQML/WITSML/PRODML-object that has
    connected array data.


    Parameters
    ----------
    epc_hdf_filename: str | pathlib.Path
        The filename of the HDF5-file that is connected to an EPC-file. This is
        typically the same filename as the EPC-file, but with the
        `.epc`-extension replaced by `.h5`.


    Returns
    -------
    dict[str | npt.NDArray[np.bool_ | np.number | np.character]
        A dictionary mapping the `pathInHdfFile`-key to the actual array as a
        NumPy-array.


    See Also
    --------
    get_resqml_v201_objects
    """
    datasets = {}

    def populate_datasets(name: str, obj: h5py.Dataset | h5py.Group) -> None:
        if isinstance(obj, h5py.Dataset):
            # The name does not include the top level group "/". This creates a
            # mismatch between the PathInHdfFile-element in the RESQML
            # documents.
            datasets["/" + name] = np.array(obj)

    with h5py.File(epc_hdf_filename, "r") as f:
        f.visititems(populate_datasets)

    return datasets



import warnings

warnings.warn(
    "The 'pyetp.types' module is deprecated and will be removed in a future "
    "version of pyetp. Please import the ETP types directly from 'etyptypes' instead.",
    DeprecationWarning,
    stacklevel=2,
)


from etptypes import ETPModel
from etptypes.energistics.etp.v12.datatypes.any_array import AnyArray
from etptypes.energistics.etp.v12.datatypes.any_array_type import AnyArrayType
from etptypes.energistics.etp.v12.datatypes.any_logical_array_type import (
    AnyLogicalArrayType,
)
from etptypes.energistics.etp.v12.datatypes.array_of_boolean import ArrayOfBoolean
from etptypes.energistics.etp.v12.datatypes.array_of_double import ArrayOfDouble
from etptypes.energistics.etp.v12.datatypes.array_of_float import ArrayOfFloat
from etptypes.energistics.etp.v12.datatypes.array_of_int import ArrayOfInt
from etptypes.energistics.etp.v12.datatypes.array_of_long import ArrayOfLong
from etptypes.energistics.etp.v12.datatypes.data_array_types.data_array import DataArray
from etptypes.energistics.etp.v12.datatypes.data_array_types.data_array_identifier import (
    DataArrayIdentifier,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.data_array_metadata import (
    DataArrayMetadata,
)
from etptypes.energistics.etp.v12.datatypes.data_value import DataValue
from etptypes.energistics.etp.v12.datatypes.error_info import ErrorInfo
from etptypes.energistics.etp.v12.datatypes.message_header import MessageHeader
from etptypes.energistics.etp.v12.datatypes.object.data_object import DataObject
from etptypes.energistics.etp.v12.datatypes.object.dataspace import Dataspace
from etptypes.energistics.etp.v12.datatypes.server_capabilities import (
    ServerCapabilities,
)
from etptypes.energistics.etp.v12.datatypes.uuid import Uuid
from etptypes.energistics.etp.v12.datatypes.supported_data_object import (
    SupportedDataObject,
)
from etptypes.energistics.etp.v12.datatypes.supported_protocol import SupportedProtocol
from etptypes.energistics.etp.v12.datatypes.version import Version
from etptypes.energistics.etp.v12.protocol.core.acknowledge import Acknowledge
from etptypes.energistics.etp.v12.protocol.core.authorize import Authorize
from etptypes.energistics.etp.v12.protocol.core.authorize_response import (
    AuthorizeResponse,
)
from etptypes.energistics.etp.v12.protocol.core.close_session import CloseSession
from etptypes.energistics.etp.v12.protocol.core.open_session import OpenSession
from etptypes.energistics.etp.v12.protocol.core.protocol_exception import (
    ProtocolException,
)
from etptypes.energistics.etp.v12.protocol.core.request_session import RequestSession
from etptypes.energistics.etp.v12.protocol.transaction.start_transaction import (
    StartTransaction,
)
from etptypes.energistics.etp.v12.protocol.transaction.commit_transaction import (
    CommitTransaction,
)
from etptypes.energistics.etp.v12.protocol.transaction.rollback_transaction import (
    RollbackTransaction,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.put_data_subarrays_type import (
    PutDataSubarraysType,
)
from etptypes.energistics.etp.v12.datatypes.object.relationship_kind import (
    RelationshipKind,
)
from etptypes.energistics.etp.v12.datatypes.object.context_info import ContextInfo
from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays import (
    PutDataSubarrays,
)
from etptypes.energistics.etp.v12.datatypes.object.context_scope_kind import (
    ContextScopeKind,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_subarrays_response import (
    PutDataSubarraysResponse,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.get_data_subarrays_type import (
    GetDataSubarraysType,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays import (
    GetDataSubarrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_subarrays_response import (
    GetDataSubarraysResponse,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.put_data_arrays_type import (
    PutDataArraysType,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays import (
    PutDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_data_arrays_response import (
    PutDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays import (
    GetDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_arrays_response import (
    GetDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata import (
    GetDataArrayMetadata,
)
from etptypes.energistics.etp.v12.protocol.data_array.get_data_array_metadata_response import (
    GetDataArrayMetadataResponse,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.put_uninitialized_data_array_type import (
    PutUninitializedDataArrayType,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays import (
    PutUninitializedDataArrays,
)
from etptypes.energistics.etp.v12.protocol.data_array.put_uninitialized_data_arrays_response import (
    PutUninitializedDataArraysResponse,
)
from etptypes.energistics.etp.v12.protocol.store.delete_data_objects import (
    DeleteDataObjects,
)
from etptypes.energistics.etp.v12.protocol.store.delete_data_objects_response import (
    DeleteDataObjectsResponse,
)
from etptypes.energistics.etp.v12.datatypes.object.resource import Resource
from etptypes.energistics.etp.v12.protocol.discovery.get_resources import GetResources
from etptypes.energistics.etp.v12.protocol.store.put_data_objects import PutDataObjects
from etptypes.energistics.etp.v12.protocol.store.put_data_objects_response import (
    PutDataObjectsResponse,
)
from etptypes.energistics.etp.v12.protocol.store.get_data_objects import GetDataObjects
from etptypes.energistics.etp.v12.protocol.store.get_data_objects_response import (
    GetDataObjectsResponse,
)
from etptypes.energistics.etp.v12.protocol.dataspace.delete_dataspaces import (
    DeleteDataspaces,
)
from etptypes.energistics.etp.v12.protocol.dataspace.delete_dataspaces_response import (
    DeleteDataspacesResponse,
)
from etptypes.energistics.etp.v12.protocol.dataspace.put_dataspaces import PutDataspaces
from etptypes.energistics.etp.v12.protocol.dataspace.put_dataspaces_response import (
    PutDataspacesResponse,
)
from etptypes.energistics.etp.v12.protocol.core.authorize import Authorize
from etptypes.energistics.etp.v12.protocol.core.authorize_response import (
    AuthorizeResponse,
)
#
# NOTE we want to `from etptypes.energistics.etp.v12 import datatypes, protocol` and use this
# however this not supported with pylance as of yet, hence so many imports are reexported here
#



import datetime
import typing as T
from uuid import uuid4

from xsdata.models.datatype import XmlDateTime

import resqml_objects.v201 as ro
from pyetp.config import SETTINGS
from resqml_objects.v201.utils import (
    common_schema_version,
    get_data_object_reference,
    resqml_schema_version,
)

if T.TYPE_CHECKING:
    from xtgeo import RegularSurface


def create_common_citation(title: str):
    return ro.Citation(
        title=title,
        creation=XmlDateTime.from_string(
            datetime.datetime.now().strftime("%Y-%m-%dT%H:%M:%S%z")
        ),
        originator=SETTINGS.application_name,
        format=f"{SETTINGS.application_name}:v{SETTINGS.application_version}",
    )


def create_common_crs(
    title: str,
    projected_epsg,
    rotation: float = 0.0,
    resqml_schema_version: str = resqml_schema_version,
):
    return ro.LocalDepth3dCrs(
        citation=create_common_citation(f"CRS for {title}"),
        schema_version=resqml_schema_version,
        uuid=str(uuid4()),
        # NOTE: I assume that we let the CRS have no offset, and add any offset
        # in the grid instead.
        xoffset=0.0,
        yoffset=0.0,
        zoffset=0.0,
        areal_rotation=ro.PlaneAngleMeasure(
            # Here rotation should be zero!
            value=rotation,
            uom=ro.PlaneAngleUom.DEGA,
        ),
        # NOTE: Verify that this is the projected axis order
        projected_axis_order=ro.AxisOrder2d.EASTING_NORTHING,
        projected_uom=ro.LengthUom.M,
        vertical_uom=ro.LengthUom.M,
        zincreasing_downward=True,
        vertical_crs=ro.VerticalCrsEpsgCode(epsg_code=projected_epsg),
        projected_crs=ro.ProjectedCrsEpsgCode(
            epsg_code=projected_epsg,
        ),
    )


def create_epc(common_schema_version: str = common_schema_version):
    return ro.EpcExternalPartReference(
        citation=create_common_citation("Hdf Proxy"),
        schema_version=common_schema_version,
        uuid=str(uuid4()),
        mime_type="application/x-hdf5",
    )


def parse_xtgeo_surface_to_resqml_grid(surf: "RegularSurface", projected_epsg: int):
    # Build the RESQML-objects "manually" from the generated dataclasses.
    # Their content is described also in the RESQML v2.0.1 standard that is
    # available for download here:
    # https://publications.opengroup.org/standards/energistics-standards/v231a

    title = surf.name or "regularsurface"

    # NOTE: xtgeo uses nrow for axis 1 in the array, and ncol for axis 0.  This
    # means that surf.nrow is the fastest changing axis, and surf.ncol the
    # slowest changing axis, and we have surf.values.shape == (surf.ncol,
    # surf.nrow). The author of this note finds that confusing, but such is
    # life.
    epc, crs, gri = instantiate_resqml_grid(
        title,
        surf.get_rotation(),
        surf.xori,
        surf.yori,
        surf.xinc,
        surf.yinc,
        surf.ncol,
        surf.nrow,
        projected_epsg,
    )
    return epc, crs, gri


def instantiate_resqml_grid(
    name: str,
    rotation: float,
    x0: float,
    y0: float,
    dx: float,
    dy: float,
    nx: int,
    ny: int,
    epsg: int,
    resqml_schema_version: str = resqml_schema_version,
):
    epc = create_epc()
    crs = create_common_crs(name, epsg, rotation)

    gri = ro.Grid2dRepresentation(
        uuid=(grid_uuid := str(uuid4())),
        schema_version=resqml_schema_version,
        surface_role=ro.SurfaceRole.MAP,
        citation=create_common_citation(name),
        grid2d_patch=ro.Grid2dPatch(
            # TODO: Perhaps we can use this for tiling?
            patch_index=0,
            # NumPy-arrays are C-ordered, meaning that the last index is
            # the index that changes most rapidly. However, xtgeo uses nrow for
            # axis 1 in the array, and ncol for axis 0. This means that
            # surf.nrow is the fastest changing axis, and surf.ncol the slowest
            # changing axis (as surf.values.shape == (surf.ncol, surf.nrow))
            fastest_axis_count=ny,
            slowest_axis_count=nx,
            geometry=ro.PointGeometry(
                local_crs=get_data_object_reference(crs),
                points=ro.Point3dZValueArray(
                    supporting_geometry=ro.Point3dLatticeArray(
                        origin=ro.Point3d(
                            coordinate1=x0,
                            coordinate2=y0,
                            coordinate3=0.0,
                        ),
                        # NOTE: The ordering in the offset-list should be
                        # preserved when the data is passed back and forth.
                        # However, _we_ need to ensure a consistent ordering
                        # for ourselves. In this setup I have set the slowest
                        # axis to come first, i.e., the x-axis or axis 0 in
                        # NumPy. The reason is so that it corresponds with the
                        # origin above where "coordinate1" is set to be the
                        # x0-coordinate, and "coordinate2" the y0-coordinate.
                        # However, we can change this as we see fit.
                        offset=[
                            # Offset for x-direction, i.e., the slowest axis
                            ro.Point3dOffset(
                                offset=ro.Point3d(
                                    coordinate1=1.0,
                                    coordinate2=0.0,
                                    coordinate3=0.0,
                                ),
                                spacing=ro.DoubleConstantArray(
                                    value=dx,
                                    count=nx - 1,
                                ),
                            ),
                            # Offset for y-direction, i.e., the fastest axis
                            ro.Point3dOffset(
                                offset=ro.Point3d(
                                    coordinate1=0.0,
                                    coordinate2=1.0,
                                    coordinate3=0.0,
                                ),
                                spacing=ro.DoubleConstantArray(
                                    value=dy,
                                    count=ny - 1,
                                ),
                            ),
                        ],
                    ),
                    zvalues=ro.DoubleHdf5Array(
                        values=ro.Hdf5Dataset(
                            path_in_hdf_file=f"/RESQML/{grid_uuid}/zvalues",
                            hdf_proxy=get_data_object_reference(epc),
                        ),
                    ),
                ),
            ),
        ),
    )
    return epc, crs, gri



import typing as T

import numpy as np
import numpy.typing as npt
from etptypes.energistics.etp.v12.datatypes.any_array import AnyArray
from etptypes.energistics.etp.v12.datatypes.any_array_type import AnyArrayType
from etptypes.energistics.etp.v12.datatypes.any_logical_array_type import (
    AnyLogicalArrayType,
)
from etptypes.energistics.etp.v12.datatypes.array_of_boolean import ArrayOfBoolean
from etptypes.energistics.etp.v12.datatypes.array_of_double import ArrayOfDouble
from etptypes.energistics.etp.v12.datatypes.array_of_float import ArrayOfFloat
from etptypes.energistics.etp.v12.datatypes.array_of_int import ArrayOfInt
from etptypes.energistics.etp.v12.datatypes.array_of_long import ArrayOfLong
from etptypes.energistics.etp.v12.datatypes.data_array_types.data_array import (
    DataArray,
)
from etptypes.energistics.etp.v12.datatypes.data_array_types.data_array_metadata import (
    DataArrayMetadata,
)

SUPPORTED_ARRAY_TYPES: T.TypeAlias = (
    ArrayOfFloat | ArrayOfBoolean | ArrayOfInt | ArrayOfLong | ArrayOfDouble
)


# See section 13.2.2.1 for the allowed mapping between logical array types and
# transport array types.
# NOTE: Currently the logical-array-mapping does not work on the
# open-etp-server. We write the relevant logical array type, but we only get
# AnyLogicalArrayType.ARRAY_OF_BOOLEAN in return from the server.
_ANY_LOGICAL_ARRAY_TYPE_MAP: dict[npt.DTypeLike, AnyLogicalArrayType] = {
    np.dtype(np.bool_): AnyLogicalArrayType.ARRAY_OF_BOOLEAN,
    np.dtype(np.int8): AnyLogicalArrayType.ARRAY_OF_INT8,
    np.dtype(np.uint8): AnyLogicalArrayType.ARRAY_OF_UINT8,
    np.dtype("<i2"): AnyLogicalArrayType.ARRAY_OF_INT16_LE,
    np.dtype("<i4"): AnyLogicalArrayType.ARRAY_OF_INT32_LE,
    np.dtype("<i8"): AnyLogicalArrayType.ARRAY_OF_INT64_LE,
    np.dtype("<u2"): AnyLogicalArrayType.ARRAY_OF_UINT16_LE,
    np.dtype("<u4"): AnyLogicalArrayType.ARRAY_OF_UINT32_LE,
    np.dtype("<u8"): AnyLogicalArrayType.ARRAY_OF_UINT64_LE,
    np.dtype("<f4"): AnyLogicalArrayType.ARRAY_OF_FLOAT32_LE,
    np.dtype("<f8"): AnyLogicalArrayType.ARRAY_OF_DOUBLE64_LE,
    np.dtype(">i2"): AnyLogicalArrayType.ARRAY_OF_INT16_BE,
    np.dtype(">i4"): AnyLogicalArrayType.ARRAY_OF_INT32_BE,
    np.dtype(">i8"): AnyLogicalArrayType.ARRAY_OF_INT64_BE,
    np.dtype(">u2"): AnyLogicalArrayType.ARRAY_OF_UINT16_BE,
    np.dtype(">u4"): AnyLogicalArrayType.ARRAY_OF_UINT32_BE,
    np.dtype(">u8"): AnyLogicalArrayType.ARRAY_OF_UINT64_BE,
    np.dtype(">f4"): AnyLogicalArrayType.ARRAY_OF_FLOAT32_BE,
    np.dtype(">f8"): AnyLogicalArrayType.ARRAY_OF_DOUBLE64_BE,
}

_INV_ANY_LOGICAL_ARRAY_TYPE_MAP: dict[AnyLogicalArrayType, npt.DTypeLike] = {
    v: k for k, v in _ANY_LOGICAL_ARRAY_TYPE_MAP.items()
}


# TODO: This map should be used once the logical-array-type is supported in the
# open-etp-server.
# _ANY_ARRAY_TYPE_MAP: dict[npt.DTypeLike, AnyArrayType] = {
#     np.dtype(np.bool_): AnyArrayType.ARRAY_OF_BOOLEAN,
#     np.dtype(np.int8): AnyArrayType.BYTES,
#     np.dtype(np.uint8): AnyArrayType.BYTES,
#     np.dtype("<i2"): AnyArrayType.BYTES,
#     np.dtype("<i4"): AnyArrayType.BYTES,
#     np.dtype("<i8"): AnyArrayType.BYTES,
#     np.dtype("<u2"): AnyArrayType.BYTES,
#     np.dtype("<u4"): AnyArrayType.BYTES,
#     np.dtype("<u8"): AnyArrayType.BYTES,
#     np.dtype("<f4"): AnyArrayType.ARRAY_OF_FLOAT,
#     np.dtype("<f8"): AnyArrayType.ARRAY_OF_DOUBLE,
#     np.dtype(">i2"): AnyArrayType.BYTES,
#     np.dtype(">i4"): AnyArrayType.BYTES,
#     np.dtype(">i8"): AnyArrayType.BYTES,
#     np.dtype(">u2"): AnyArrayType.BYTES,
#     np.dtype(">u4"): AnyArrayType.BYTES,
#     np.dtype(">u8"): AnyArrayType.BYTES,
#     np.dtype(">f4"): AnyArrayType.BYTES,
#     np.dtype(">f8"): AnyArrayType.BYTES,
# }


# This AnyArrayType-map is used until the logical-array-type is properly
# implemented for the open-etp-server. In this case we
_ANY_ARRAY_TYPE_MAP: dict[npt.DTypeLike, AnyArrayType] = {
    np.dtype(np.bool_): AnyArrayType.ARRAY_OF_BOOLEAN,
    np.dtype(np.int8): AnyArrayType.BYTES,
    np.dtype("<i4"): AnyArrayType.ARRAY_OF_INT,
    np.dtype("<i8"): AnyArrayType.ARRAY_OF_LONG,
    np.dtype("<f4"): AnyArrayType.ARRAY_OF_FLOAT,
    np.dtype("<f8"): AnyArrayType.ARRAY_OF_DOUBLE,
}
valid_dtypes = list(_ANY_ARRAY_TYPE_MAP)

_INV_ANY_ARRAY_TYPE_MAP: dict[AnyArrayType, npt.DTypeLike] = {
    AnyArrayType.ARRAY_OF_BOOLEAN: np.dtype(np.bool_),
    # The BYTES-arrays are converted to the proper dtype using the logical
    # array type. We can therefore interpret the bytes as np.int8, before we
    # combine the byte strings to the proper type.
    AnyArrayType.BYTES: np.dtype(np.int8),
    AnyArrayType.ARRAY_OF_INT: np.dtype("<i4"),
    AnyArrayType.ARRAY_OF_LONG: np.dtype("<i8"),
    AnyArrayType.ARRAY_OF_FLOAT: np.dtype("<f4"),
    AnyArrayType.ARRAY_OF_DOUBLE: np.dtype("<f8"),
}


_ANY_ARRAY_MAP: dict[AnyArrayType, SUPPORTED_ARRAY_TYPES] = {
    AnyArrayType.ARRAY_OF_FLOAT: ArrayOfFloat,
    AnyArrayType.ARRAY_OF_DOUBLE: ArrayOfDouble,
    AnyArrayType.ARRAY_OF_INT: ArrayOfInt,
    AnyArrayType.ARRAY_OF_LONG: ArrayOfLong,
    AnyArrayType.ARRAY_OF_BOOLEAN: ArrayOfBoolean,
    AnyArrayType.BYTES: bytes,
}

_INV_ANY_ARRAY_MAP: dict[SUPPORTED_ARRAY_TYPES, AnyArrayType] = {
    v: k for k, v in _ANY_ARRAY_MAP.items()
}


def check_if_array_is_valid_dtype(array: npt.NDArray[T.Any]) -> bool:
    return array.dtype in valid_dtypes


def get_valid_dtype_cast(array: npt.NDArray[T.Any]) -> npt.DTypeLike:
    if check_if_array_is_valid_dtype(array):
        return array.dtype

    if array.dtype == np.dtype(np.uint8):
        return np.dtype(np.int8)
    elif array.dtype == np.dtype("<u2"):
        return np.dtype("<i2")
    elif array.dtype == np.dtype(">u2"):
        return np.dtype("<i2")
    elif array.dtype == np.dtype("<u4"):
        return np.dtype("<i4")
    elif array.dtype == np.dtype(">u4"):
        return np.dtype("<i4")
    elif array.dtype == np.dtype("<u8"):
        return np.dtype("<i8")
    elif array.dtype == np.dtype(">u8"):
        return np.dtype("<i8")

    raise TypeError(f"Dtype {array.dtype} does not have a valid cast")


def get_logical_array_type(dtype: npt.DTypeLike) -> AnyLogicalArrayType:
    logical_array_type = _ANY_LOGICAL_ARRAY_TYPE_MAP.get(dtype)

    if logical_array_type is not None:
        return logical_array_type

    # Here we might be taking a chance by not caring about the endianess of the
    # string.
    if dtype.type == np.str_:
        return AnyLogicalArrayType.ARRAY_OF_STRING

    # We ignore the AnyLogicalArrayType.ARRAY_OF_CUSTOM for now.
    raise KeyError(f"Data type {dtype} is not a valid ETP v1.2 logical array type")


def get_transport_array_type(dtype: npt.DTypeLike) -> AnyArrayType:
    transport_array_type = _ANY_ARRAY_TYPE_MAP.get(dtype)

    if transport_array_type is not None:
        return transport_array_type

    # Here we might be taking a chance by not caring about the endianess of the
    # string.
    if dtype.type == np.str_:
        return AnyArrayType.ARRAY_OF_STRING

    raise KeyError(
        f"Data type {dtype} does not have a valid map to an ETP v1.2 transport array "
        f"type. Valid types are: {list(_ANY_ARRAY_TYPE_MAP)}"
    )


def get_logical_and_transport_array_types(
    dtype: npt.DTypeLike,
) -> tuple[AnyLogicalArrayType, AnyArrayType]:
    # See section 13.2.2.1 in the ETP v1.2 specification for the allowed
    # mappings between the logical and transport types.
    # Using this function ensures that the combination of the logical and
    # transport array types are valid (it is set up in valid combinations in
    # the mapping dictionaries at the top).

    return get_logical_array_type(dtype), get_transport_array_type(dtype)


def get_etp_data_array_from_numpy(data: npt.NDArray) -> DataArray:
    transport_array_type = get_transport_array_type(data.dtype)
    cls = _ANY_ARRAY_MAP[transport_array_type]

    if cls is bytes:
        # In the current implementation we only support 1-byte sized dtype's
        # when using "bytes". In the future, with logical array types, this can
        # cover multiple dtypes, but then the dimensions must be adjusted.
        itemsize = data.dtype.itemsize
        item = np.ravel(data).tobytes()
        dimensions = list(data.shape)
        dimensions[-1] = dimensions[-1] * itemsize

        return DataArray(dimensions=dimensions, data=AnyArray(item=item))

    return DataArray(
        dimensions=data.shape, data=AnyArray(item=cls(values=np.ravel(data).tolist()))
    )


def get_transport_array_size(metadata: DataArrayMetadata) -> int:
    dtype = _INV_ANY_ARRAY_TYPE_MAP[metadata.transport_array_type]
    return int(np.prod(metadata.dimensions) * dtype.itemsize)


def get_dtype_from_any_array_class(cls: AnyArray) -> npt.DTypeLike:
    if cls is bytes:
        return np.dtype(np.int8)
    elif cls is ArrayOfBoolean:
        return np.dtype(np.bool_)
    elif cls is ArrayOfInt:
        return np.dtype("<i4")
    elif cls is ArrayOfLong:
        return np.dtype("<i8")
    elif cls is ArrayOfFloat:
        return np.dtype("<f4")
    elif cls is ArrayOfDouble:
        return np.dtype("<f8")
    # TODO: Update NumPy to >= 2.0, and import the ArrayOfString-class
    # elif cls == ArrayOfString:
    #     return np.StringDType()

    raise TypeError(f"Class {cls} is not a valid array class")


def get_dtype_from_any_array_type(_type: T.Union[AnyArrayType | str]) -> npt.DTypeLike:
    enum_name = AnyArrayType(_type)
    return _INV_ANY_ARRAY_TYPE_MAP[enum_name]


def get_numpy_array_from_etp_data_array(
    data_array: DataArray,
) -> npt.NDArray[
    # The types used here do not tell which endianess is used for the returned
    # arrays, but until we can use np.dtype("<f4")-like syntax (Python > 3.10),
    # this will do.
    np.int8 | np.bool_ | np.int32 | np.int64 | np.float32 | np.float64
]:
    dtype = get_dtype_from_any_array_class(type(data_array.data.item))

    if type(data_array.data.item) is bytes:
        return np.array(np.frombuffer(data_array.data.item, dtype=dtype)).reshape(
            data_array.dimensions
        )

    return np.array(data_array.data.item.values, dtype=dtype).reshape(
        data_array.dimensions
    )



from lxml import etree
from xsdata.formats.dataclass.context import XmlContext
from xsdata.formats.dataclass.parsers import XmlParser

import resqml_objects.v201 as ro_201


def parse_resqml_v201_object(raw_data: bytes) -> ro_201.AbstractObject:
    parser = XmlParser(context=XmlContext())

    obj_name = etree.QName(etree.fromstring(raw_data).tag).localname
    return parser.from_bytes(raw_data, getattr(ro_201, obj_name))



from ._version import __version__
from .client import ETPClient, ETPError, connect
from .uri import DataObjectURI, DataspaceURI

__all__ = [
    "__version__",
    "ETPClient",
    "ETPError",
    "connect",
    "DataObjectURI",
    "DataspaceURI",
]



from xsdata.formats.dataclass.serializers import XmlSerializer
from xsdata.formats.dataclass.serializers.config import SerializerConfig

import resqml_objects.v201 as ro_201


def serialize_resqml_v201_object(obj: ro_201.AbstractObject) -> bytes:
    serializer = XmlSerializer(config=SerializerConfig())

    return str.encode(serializer.render(obj))
